# $Id: Legato.pm,v 1.3.8.3 2001/06/13 00:33:49 bservies Exp $
#
# Legato.pm
# Legato related routines
#
# Copyright 2001 Sun Microsystems, Inc.
# All Rights Reserved
#

package Cobalt::Legato;

use vars qw(@ISA @EXPORT @EXPORT_OK);
require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT	= qw(
	set_state
	sync_ports
	set_config
	get_config
	set_ports
	get_ports
	set_servers
	get_servers
);

use Cobalt::Product;
use Cobalt::Netutil;
use Cobalt::Services;
use Cobalt::Util;

1;
__DATA__

use strict;

sub set_state
# Description: restart the NetWorker backup client daemon
# Inputs: state (0=off, 1=on, 2=restart)
# Outputs: MSG_ok
# Side effects: fork to restart the service.
{
	my ($state) = @_;
	my $pid;
	my $script = "/etc/rc.d/init.d/$NetWorker_init_script";

	# Determine what action to take on the networker startup script.
	if ($state == 0) {
		$action = "stop";
	} elsif ($state == 1) {
		$action = "start";
	} elsif ($state == 2) {
		$action = "restart";
	} else {
		return MSG_error("services_networker_setstate_invalid", $state);
	}

	if (!defined($pid = fork)) {
		die "cannot fork: $!";
	} elsif ($pid) {
		# parent process
		print STDERR "Forking to restart NetWorker.\n" if ($Debug);
		return MSG_ok("services_networker_setstate_ok");
	} else {
		# child process

		# First, restart the NetWorker client daemon
		system("$script $action >> $Sauce_log 2>&1") &&
		    die MSG_error("services_networker_setstate_failed", $!);

		# Second, apply any port range settings if we started or
		# restarted NetWorker.
		if ($state) {
			sync_ports();
		}

		# end the child process
		exit;
	}
}

sub sync_ports
# Description: Syncronize the port ranges specified in the UI with the
# running NetWorker client.
# Inputs: none
# Outputs: none
# Side effects: runs nsrports to reset port ranges.
{
	my ($s_ports, $c_ports);
	my $nsrports = "/usr/bin/nsrports";

	# Get the service and connection port ranges
	($s_ports, $c_ports) = get_ports();

	# Run the command to reset the port ranges.
	system("$nsrports -S $s_ports >> $Sauce_log 2>&1") && return;
	system("$nsrports -C $c_ports >> $Sauce_log 2>&1") && return;
}


sub set_config
# Description: configure NetWorker using the given information
# Inputs: service port range, connection port range, list of servers
# Outputs: MSG_ok or MSG_error
{
	my ($s_ports, $c_ports, @serverlist) = @_;
	my $result;

	# Configure the backup servers
	$result = set_servers(@serverlist);
	if ($result !~ /^2/) {
		# An error occurred.  Return the message
		print 
		return ($result);
	}

	# Configure the service port range
	$result = set_ports($s_ports, $c_ports);
	if ($result !~ /^2/) {
		# An error occurred.  Return the message
		return ($result);
	}
	
	if (Cobalt::Services::service_get_active("networker")) {
		# Restart NetWorker for changes to take effect
		$result = set_state(2);
		if ($result !~ /^2/) {
			# An error occurred.  Return the message
			return ($result);
		}
	}

	return MSG_ok("services_networker_configure_ok");
}


sub get_config
# Description: get the configuration of NetWorker
# Inputs: none
# Outputs: service port range, connection port range, list of servers
{
	my ($s_ports, $c_ports, @serverlist);

	# Get the configuration information
	($s_ports, $c_ports) = get_ports();
	@serverlist = get_servers();

	return ($s_ports, $c_ports, @serverlist);
}


sub set_ports
# Description: configure a port range for networker
# Inputs: service port range, connection port range
# Outputs: MSG_ok or MSG_error
{
	my ($s_ports, $c_ports) = @_;
	my ($low, $high, $block_data, $result);

	# Validate the service port range.
	($low, $high) = ($s_ports =~ /\s*(\d+)\s*-\s*(\d+)\s*/);
	if ($low >= $high) {
		# Invalid port range. Low must be less than high
		return MSG_error("services_networker_portrange_invalid",
		    $low, $high);
	}
	$s_ports = $low . "-" . $high;

	# Validate the connection port range.
	($low, $high) = ($c_ports =~ /\s*(\d+)\s*-\s*(\d+)\s*/);
	if ($low >= $high) {
		# Invalid port range. Low must be less than high
		return MSG_error("services_networker_portrange_invalid",
		    $low, $high);
	}
	$c_ports = $low . "-" . $high;

	# Create the data block to store
	$block_data = "";
	$block_data .= "Service ports: " . $s_ports . "\n";
	$block_data .= "Connection ports: " . $c_ports . "\n";

	#
	# Store the port range in a file.  Port ranges can only be updated
	# when nsrexecd is running (the NetWorker client daemon), so we must
	# cache it in case the service is not currently enabled.  The port
	# ranges are applied when the service is started.
	#
	$result = Cobalt::Util::replaceblock($NetWorker_port_file,
	    $NetWorker_config_start, $block_data, $NetWorker_config_end,
	    $NetWorker_config_mode);
	if ($result == 0) {
		# And error occured storing the port data
		return MSG_error("services_networker_configure_failed", $!);
	}

	return MSG_ok("services_networker_configure_ok");
}


sub get_ports
# Description: Get a range of ports from NetWorker
# Inputs: none
# Outputs: service ports, connection ports
{
	my ($in_block, $result, $s_ports, $c_ports);

	if (-e $NetWorker_port_file) {
		# The port range cache exists, so lets read from it
		Cobalt::Util::FILE_lock_r($NetWorker_port_file, *IN);
	} else {
		#
		# The port range cache does not exist.  Return some default
		# ranges.  We cannot read from the NetWorker client defaults
		# because the service may not be running.
		#
		return ("9000-9010", "9011-9999");
	}

	$in_block = 0;
	while (<IN>) {
		if (/^$NetWorker_config_start/) {
			$in_block = 1;
		}
		next unless $in_block;

		if (/^$NetWorker_config_end/) {
			$in_block = 0;
			next;
		}

		if (/^Service ports:\s*([\d-]+)/) {
			$s_ports = $1;
		}
		if (/^Connection ports:\s*([\d-]+)/) {
			$c_ports = $1;
		}
	}
	close (IN);

	Cobalt::Util::FILE_unlock($NetWorker_port_file);

	return ($s_ports, $c_ports);
}


sub set_servers
# Description: configure the list of servers used by NetWorker
# Inputs: list of servers
# Outputs: MSG_ok or MSG_error
{
	my @serverlist = @_;
	my ($server, $block_data, $result);

	$block_data = "";
	foreach $server (@serverlist) {
		# Make sure the server ip can be determined
		$result = Cobalt::Netutil::netutil_dns_verifyhost($server);
		if ($result !~ /^2/) {
			return ($result);
		}
		$block_data .= "$server\n";
	}

	$result = Cobalt::Util::replaceblock($NetWorker_server_file,
	    $NetWorker_config_start, $block_data, $NetWorker_config_end,
	    $NetWorker_config_mode);
	if ($result == 0) {
		# And error occured storing the port data
		return MSG_error("services_networker_configure_failed", $!);
	}

	return MSG_ok("services_networker_configure_ok");
}


sub get_servers
# Description: get the list of servers used by the NetWorker client
# Inputs: none
# Outputs: list of servers (may be empty)
{
	my @serverlist = ();

	if (! -e $NetWorker_server_file) {
		# Return the empty list since the server list file does not
		# exist
		return @serverlist;
	}

	# Lock the file and open it for reading
	Cobalt::Util::FILE_lock_r($NetWorker_server_file, *IN);
	while (<IN>) {
		chop; # Remove trailing newline
		next if (/^#/);	# skip comments
		next if (/^\s*$/); # skip empty lines

		# The only valid thing on the line is a single server entry
		# Add it to the list
		push @serverlist, ($_);
	}
	close (IN);

	# Unlock the file
	Cobalt::Util::FILE_unlock($NetWorker_server_file);

	return (@serverlist);
}

1;

