# $Id: NetBackup.pm,v 1.1.2.4 2001/06/13 00:36:16 bservies Exp $
#
# Copyright 2001 Sun Microsystems, Inc.  All rights reserved.
#
# NetBackup.pm
# NetBackup related routines
#

package Cobalt::NetBackup;

use vars qw(@ISA @EXPORT @EXPORT_OK);
require Exporter;
require SelfLoader;
@ISA    = qw(Exporter SelfLoader);
@EXPORT	= qw(
		get_servers
		get_state
		set_state
		configure
	    );

use Cobalt::Product;
use Cobalt::Network;
use Cobalt::Netutil;
use Cobalt::Util;

1;

__DATA__

use strict;

sub configure
# Description:
#  Inputs: list of servers to listen two
# Outputs: NetBackup configuration file updated.  
{
	my (@serverlist) = @_;
	my $block_data = "";
	my $result;
	my ($ip, $mask, $hwaddr);
	my $hostname = Cobalt::Network::net_get_hostname();

	# Build the configuration block to store
	foreach $server (@serverlist) {
		# Verify the hostname
		$result = Cobalt::Netutil::netutil_dns_verifyhost($server);
		if ($result !~ /^2/) {
			return ($result);
		}
		$block_data .= "SERVER = $server\n";
	}

	#
	# After talking with Veritas, the general practice for the CLIENT_NAME
	# parameter seems to be the hostname, rather than the FQDN.
	#
	$block_data .= "CLIENT_NAME = $hostname\n";

	# Replace the data block in the configuration file.
	$result = Cobalt::Util::replaceblock($NetBackup_config_file,
	    $NetBackup_config_start, $block_data, $NetBackup_config_end,
	    $NetBackup_config_mode);
	if ($result == 0) {
		# An error occurred setting the server information
		return (MSG_error("services_netbackup_enable_failed", $!));
	}

	# It is not necessary to restart NetBackup, since inetd launches it
	# for each connection.
	return MSG_ok("services_netbackup_configure_ok");
}

sub set_state
# Description: Enables or disables NetBackup client
# Inputs: state = 1 (on) or 0 (off)
# Outputs: 
{
	my ($state) = @_;
	my $inetd_data;
	my $inetd_pid;
	my $result;
	my $services_data;

	if (! $Features{'netbackup'}) {
		return MSG_error("services_netbackup_unsupported");
	}

	if ($state == 0) {
		# Turn NetBackup off
		$inetd_data = "";
		$services_data = "";
	} else {
		# Turn NetBackup on

		$inetd_data = "";
		$inetd_data .= "bpcd    stream  tcp     nowait  root    /usr/openv/netbackup/bin/bpcd   bpcd\n";
		$inetd_data .= "vopied  stream  tcp     nowait  root    /usr/openv/bin/vopied vopied\n";
		$inetd_data .= "bpjava-msvc     stream  tcp     nowait  root    /usr/openv/netbackup/bin/bpjava-msvc -transient\n";                                           
		$services_data = "";
		$services_data .= "bprd    13720/tcp       bprd\n";
		$services_data .= "bpcd    13782/tcp       bpcd\n";
		$services_data .= "vopied  12783/tcp       vopied\n";
		$services_data .= "bpjava-msvc     13722/tcp       bpjava-msvc\n";
	}

	# Update the inetd configuration
	$result = replaceblock($Inetd_conf, $NetBackup_config_start,
	    $inetd_data, $NetBackup_config_end, $Inetd_mode);
	if ($result == 0) {
		return (MSG_error("services_netbackup_enable_failed", $!));
	}


	# Update the services configuration
	$result = replaceblock($Services_conf, $NetBackup_config_start,
	    $services_data, $NetBackup_config_end, $Services_mode);
	if ($result == 0) {
		return (MSG_error("services_netbackup_enable_failed", $!));
	}

	# Determine inetd's PID so we can restart it and enable netbackup
	open(PID,$Inetd_pid) || return MSG_error("3466", $!);
	while (<PID>) {
		if (/^(\d+)$/o) {
			$inetd_pid = $1;
		}
	}
	close PID;

	#
	# Restart inetd with a HUP signal if the PID was found.  If the PID was
	# not found, count this as success, since netbackup will be enabled
	# when inetd is eventually restarted (probably by swatch).
	#
	if ($inetd_pid) {
		$result = kill("HUP",$inetd_pid);
		if ($result != 1) {
			#
			# Kill returns the number of processes signaled.  If
			# no process was signaled, inetd was not restarted.
			#
			return (MSG_error("services_netbackup_enable_failed",
				$!));
		}
	}

	return MSG_ok("services_netbackup_enable_ok");
}


sub get_state
# Description: Determines if NetBackup client is enabled
# Inputs: none
# Outputs: 0 = disabled, 1 = enabled
{
	my @config_files = ($Inetd_conf, $Services_conf);
	my $bpcd = 0;
	my $bprd = 0;
	my $vopied = 0;
	my $bpjava = 0;
	my ($ifile, $result);

	#
	# Count the number of times the 4 services occur in /etc/inetd and
	# /etc/services.  There should be 2 of everything except bprd, which
	# should occur once.
	#
	foreach $ifile (@config_files) {
		open(IFILE, $ifile) ||
		    return (MSG_error("netbackup_noconf", $ifile));
		while (<IFILE>) {
			# Count the occurances of each service
			$bpcd++ if (/^bpcd/);
			$bprd++ if (/^bprd/);
			$vopied++ if (/^vopied/);
			$bpjava++ if (/^bpjava/);
		}
		close(IFILE);
	}

	if ($bpcd || $bprd || $vopied || $bpjava) {
		# It is at least partially configured ...
		$result = -1;
		if ($bpcd == 2 && $bprd == 1 && $vopied == 2 && $bpjava == 2) {
			# bprd is only set in services by the NetBackup install
			# for some reason.
			$result = 1;
		}
		if ($result < 0) {
			# Print an error and indicate that NetBackup is not
			# configured.
			print STDERR "Warning: NetBackup client partially configured!\n";
			$result = 0;
		}
	} else {
		$result = 0;
	}

	return ($result);
}

sub get_servers
{
	my @servers = ();
	open(IFILE, $NetBackup_config_file) ||
		MSG_error("netbackup_noconf", $NetBackup_config_file);
	while (<IFILE>) {
		if (/^SERVER = (\S+)/) {
			push @servers, ($1);
		}
	}
	close(IFILE);
	return @servers;
}

1;

