#!/usr/bin/perl

# useful functions... nuff said!
#
# written by Duncan Laurie <duncan@cobaltnet.com>
#

package Cobalt::Util;

use POSIX;
use Locale::gettext;

require Exporter;
require SelfLoader;
@ISA = qw(Exporter SelfLoader);
@EXPORT = qw(FILE_lock_r FILE_lock_w FILE_lock_rw
	     FILE_unlock FILE_unlock_rw
	     MSG_error MSG_abort MSG_ok MSG_get MSG_get_ui
             parse_cgipost replaceblock copy_access_bits switch_files);

use vars qw($DEBUG);

use FileHandle;
sub SAFEMODE { O_WRONLY | O_CREAT | O_EXCL };

$DEBUG=0;

1;

__DATA__

# Cobalt::Util::MSG_get_ui
#
# generate an message for the UI
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_get_ui
{
    my $msgkey = shift;
    my @vars = @_;

    my $msg = Locale::gettext::dgettext("ui", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    return $msg;
}

# Cobalt::Util::MSG_get
#
# generate an message
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_get
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    return $msg;
}

# Cobalt::Util::MSG_error
#
# generate an error message prefixed with "4999" from gettext
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_error
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '4999 '.$msg;
    return $msg;
}

# Cobalt::Util::MSG_abort
#
# generate an error message prefixed with "3999" from gettext
# using predetermined locale
#
# arguments are substituted for $arg in the message string
#
sub MSG_abort
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '3999 '.$msg;
    return $msg;
}

# Cobalt::Util::MSG_ok
#
# generate a message from the provided key using pre-defined locale
# prefix this message with "2999" to show that it is a good message
#
# arguments are substituted for $arg in the message string
#
sub MSG_ok
{
    my $msgkey = shift;
    my @vars = @_;
    my $msg = Locale::gettext::dgettext("special-sauce", $msgkey);
    foreach (@vars) {
	$msg =~ s/\$[^\\\ \.\"\']+/$_/;
    }
    $msg = '2999 '.$msg;
    return $msg;
}

# Cobalt::Util::FILE_lock_rw
#
# inputs:  filein = name of file to lock (& full path)
# return:  two filehandles, the first for the input and the second for output
#
# creates a lockfile in $Lockdir (/etc/cobalt/locks)
# for writing, while opening the $file for reading
# 
sub FILE_lock_rw
{
    my $filein = shift;
    local *FHIN = shift;
    local *FHOUT = shift;
    my $LOCK_EX = 2;
    my $Lockdir = "/etc/locks";

    # extract filename from fully-qualified filename
    my @fary = split("/", $filein);
    my $ret = pop(@fary);
    my $fileout = "$Lockdir/".pop(@fary)."_$ret";

    # we want to have exclusive locking on the input file
    print STDERR "FILE_lock_rw:  opening $filein... " if ($DEBUG);
    unless(open(FHIN, "< $filein")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  locking $filein... " if ($DEBUG);
    unless(flock(FHIN,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  opening $fileout... " if ($DEBUG);
    unless(open(FHOUT, "> $fileout")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_lock_rw:  locking $fileout... " if ($DEBUG);
    unless(flock(FHOUT,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_lock_w
#
# inputs:  file = name of file to lock (& full path)
# return:  filehandle
#
# opens the $file for reading
#
sub FILE_lock_w
{
    my $file = shift;
    local *FH = shift;
    my $LOCK_EX = 2;

    # open the file
    print STDERR "FILE_lock_w:  opening $file... " if ($DEBUG);
    unless(open(FH, "> $file")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    # lock the file
    print STDERR "FILE_lock_w:  locking $file... " if ($DEBUG);
    unless(flock(FH,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_lock_r
#
# inputs:  file = name of file to lock (& full path)
# return:  filehandle
#
# opens the $file for reading
#
sub FILE_lock_r
{
    my $file = shift;
    local *FH = shift;
    my $LOCK_EX = 2;

    # open the file
    print STDERR "FILE_lock_r:  opening $file... " if ($DEBUG);
    unless(open(*FH,"< $file")) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    # lock the file
    print STDERR "FILE_lock_r:  locking $file... " if ($DEBUG);
    unless(flock(FH,$LOCK_EX)) {
	print STDERR "error!\n" if ($DEBUG);
	return;
    }
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

#
# Cobalt::Util::FILE_unlock
#
# inputs:  array of filehandles to unlock and close
# return:  1 if success
#
sub FILE_unlock
{
    my (@fhs) = @_;
    my ($fh);
    my $LOCK_UN = 8;

    foreach $fh (@fhs) {
	print STDERR "FILE_unlock:  unlocking filehandle >$fh<... " if ($DEBUG);
	flock($fh, $LOCK_UN);
	close($fh);
	print STDERR "done\n" if ($DEBUG);
    }

    return(1);
}

#
# Cobalt::Util::FILE_unlock_rw
#
# inputs:  $filein = full path to file
#          $fhin = input filehandle returned by FILE_lock_rw
#          $fhout = output filehandle returned by FILE_lock_rw
# return:  1 if succes
#
sub FILE_unlock_rw
{
    my ($filein,$fhin,$fhout) = @_;
    my $LOCK_UN = 8;
    my $Lockdir = "/etc/locks";

    # extract filename from fully-qualified filename
    my @fary = split("/", $filein);
    my $ret = pop(@fary);
    my $fileout = "$Lockdir/".pop(@fary)."_$ret";

    print STDERR "FILE_unlock_rw:  unlocking $fileout... " if ($DEBUG);
    flock($fhout, $LOCK_UN);
    close($fhout);
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_unlock_rw:  moving $fileout to $filein... " if ($DEBUG);
    qx[ mv -f $fileout $filein ];
    print STDERR "done\n" if ($DEBUG);

    print STDERR "FILE_unlock_rw:  unlocking $filein... " if ($DEBUG);
    flock($fhin, $LOCK_UN);
    close($fhin);
    print STDERR "done\n" if ($DEBUG);

    return(1);
}

# 
# Cobalt::Util::parse_cgipost
# 
# inputs: none explicitly, obtained from environment, both POST and GET
#
# return: hash of CGI variable name key and the variable definition as target
#
sub parse_cgipost
{
    my ($k,@equality,%fields);
    my $input = '';

    read(STDIN,$input,$ENV{"CONTENT_LENGTH"}) if ($ENV{"CONTENT_LENGTH"} > 0 );
    $input .= $ENV{"QUERY_STRING"}."&";

    foreach $_ ( split(/&/,$input) )
        {
        /(.*)=(.*)/;
        my $left = $1;
        my $right = $2;

        $left =~ s/\%(.{2})/sprintf("%c",hex($1))/eg;
        $left =~ tr/+/ /s;

        $right =~ s/\%(.{2})/sprintf("%c",hex($1))/eg;
        $right =~ tr/+/ /s;
        $right =~ s/[#\$\*\\\^\(\)\n]/ /og;
        $right =~ s/\+/ /g;
        $right =~ s/  / /g;
        $right =~ s/\"/\"\"/g;

        if ($fields{"$left"})
            {
            $fields{"$left"} .= ", $right" unless ($right eq "");
            }
        else
            {
            $fields{"$left"} = $right;
            }
        }

    $fields{"remote_addr"} = $ENV{"REMOTE_ADDR"};
    $fields{"remote_host"} = $ENV{"REMOTE_HOST"};

    return(%fields);
}


# replaceblock( $filename, $start_line, $data, $end_line, $filemode );
#
# replaces the section of $filename denoted by the $start_line and
# $end_line with the contents of $data.
#
# Caveat: if $data is an empty string, the (empty) block is left in
# place in the file as a placeholder.  OTOH, if $data is undefined,
# the block (including start and stop tags) are completely removed
# from the file.
#
# Caveat: start_line and end_line must start at the beginning of a new line.
sub replaceblock
{
  my $filename = shift;
  my $start_line = shift;
  my $data = shift;
  my $end_line = shift;
  my $fmode = shift || undef;
  
  my $ret = 1;

  # attempt to handle the case where the file is missing:
  if (!-e $filename) {
    print STDERR "Warning: File $filename was missing.\n";
    unlink($filename);
    my $fh = new FileHandle("$filename", SAFEMODE);
    if ($fh) {
    	$fh->close();
    } else {
      print STDERR "Could not create $filename: $!\n";
      return 0; # fail
    }
  }

  # make a place for the new edition:
  my $newfile = $filename . '~';
  unlink($newfile) if (-e $newfile);

  # open for reading
  my $ret = FILE_lock_r($filename, *FIN);
  if (!$ret) {
    print STDERR "Could not open $filename for replacing: $!\n";
    return 0;
  }

  # open for writing
  my @statbuf = stat($filename);
  unlink($newfile);
  my $fout = new FileHandle("$newfile", SAFEMODE, $statbuf[2]);
  if (!$fout) {
    print STDERR "Could not open $newfile for writing: $!\n";
    return 0;
  }

  # precompile regexi:
  my $start_re = qr/^$start_line\s*$/;
  my $end_re = qr/^$end_line\s*$/;
 
  # process file:
  my $mode = 'init';
  while (defined($_ = <FIN>)) {
    if ($mode eq 'init') {
      # initial state (print anything but a start tag)
      if ($_ =~ $start_re) {
        # encountered start tag
        $mode = 'inside';
        if (defined($data)) {
          print $fout $start_line,"\n",$data,"\n",$end_line,"\n";
        }
        next;
      }
      print $fout $_;
      next;
    }
    if ($mode eq 'inside') {
      # inside block state (don't print anything)
      if ($_ =~ $end_re) {
        # encountered an end tag
        $mode = 'outside';
      }
      next;
    }
    if ($mode eq 'outside') {
      # past block state (print anything)
      print $fout $_;
      next;
    }
  } # end of while loop
  if ($mode eq 'init' && defined($data)) {
    # tag was never found, so let's make one:
    print $fout $start_line,"\n",$data,"\n",$end_line,"\n";
  }
    
  # close file handles
  $fout->close();
  close(FIN);
  
  # set up permissions:
  copy_access_bits($filename, $newfile);
  defined($fmode) && chmod($fmode, $newfile);

  # pull the old file switcheroo:
  switch_files($filename, $newfile);

  # unlock the file
  FILE_unlock($filename);
  
  # return success
  return $ret;
}


sub switch_files
{
  my ($f1, $f2) = (shift, shift);

  if (!-e $f1) { return -9; }
  if (!-e $f2) {
    link $f1, $f2;
    return 1;
  }
  
  my $tmp = $f1 . ".$$.tmp";

  do { link ($f1, $tmp) } || return -1;
  unlink($f1) || do {
    # undo
    unlink $tmp;
    return -2;
  };
  
  link ($f2, $f1) && unlink($f2) || do {
    # undo
    unlink($f1);
    link ($tmp, $f1) && unlink($tmp);
    return -3;
  };
  
  link ($tmp, $f2) || do {
    # undo
    link($f1, $f2);
    unlink($f1);
    link ($tmp, $f1) && unlink($tmp);
    return -4;
  };
  unlink($tmp);
  return 1;
} 
  

sub copy_access_bits
{
  my ($srcfile, $destfile) = (shift, shift);

  my @buf;
  @buf = stat($srcfile);
  if (!@buf) {
    print STDERR "Couldn't stat $srcfile\n" if ($DEBUG);
    return 0;
  }
  
  my $mode = $buf[2] & 07777;
  chmod($mode, $destfile) || do {
    print STDERR "Couldn't chmod $mode, $destfile\n" if ($DEBUG);
    return 0;
  };
  
  my $uid = $buf[4];
  my $gid = $buf[5];
  chown ($uid, $gid, $destfile) || do {
    print STDERR "Couldn't chown $destfile\n" if ($DEBUG);
    return 0;
  };
  
  return 1;
}

1;

